#lang racket

;;; missp.rkt, pour le TP L3-INFO, UNSA 2007-08, Racket
;;; TRAVAILLEZ DANS CE FICHIER !
;;; Utilisation de lex/yacc pour compiler du MISSP vers du Scheme...

;;; ------------------ construction du fichier de test ---------------------------------

(define string-for-file    ; contient un programme valide MISSP
  "foo {
         let b := false and n := 10 and id := proc(x) {x} and f := 9999 and iter := 9999
         in {letrec fac := proc(n,f) {if -n=0 then f(1) else fac(n-1,proc(r) {f(r*n)})}
              in {fac(n,id)}};
         n := 9;
         ++n;
         f := 1;
         f;
         iter := proc() { if n>0 then f else { f:=f*n; n--; iter()}};
         iter();
         print(f)
       } ")


(call-with-output-file "test.msa"        ; construction du fichier de test
    (lambda (p-out)
      (fprintf p-out "~a\n" string-for-file))
    #:exists 'replace)

;;; ------------------------------------------------------------------------------------

(require parser-tools/lex parser-tools/yacc)

(printf "Generating lexer...~n")

(define-lex-abbrevs
  [whitespace (union #\newline #\return #\tab #\space)]
  [letter ...]
  [digit ...]
  [int ...]
  [float ...]
  [boolean ...]
  [var ...])

(define-tokens value-tokens (BOOL NUM VAR))

(define-empty-tokens op-tokens
  (NEWLINE LPAR RPAR LBRA RBRA COLON SEMICOLON EOF define if then else LOWER_THAN_ELSE true false
           begin end let letrec in and NEG proc + - * / ^ = > < >= <= := -- ++))

(define get-lexeme
  (lexer
   [(eof) 'EOF]
   [(repetition 1 +inf.0 whitespace) (get-lexeme input-port)]
   ["(" 'LPAR]
   [")" ...]
   ["{" 'LBRA]
   ["}" ...]
   ["," ...]
   [";" 'SEMICOLON]
   [boolean (token-BOOL ...)]   ; on transforme true et false en #t et #f
   [(union "if" "then" "else" "let" "letrec" "define" "proc" "begin" "end" "in" "and") (string->symbol lexeme)]
   [(union "+" "-" "*" "/" "^" ">" "<" ">=" "<=" "=" ":=" "--" "++") (string->symbol lexeme)]
   [var ...]
   [int ...]
   [float ...]
   [(char-range #\000 #\377) (error "Unparseable" (string->symbol lexeme))]))

(when #f             ; mettre #t pour activer ce test !!
  (printf "Testing lexer on file \"test.msa\"...\n")
  (call-with-input-file "test.msa"
    (lambda (p-in) 
      (do ((i 0 (+ i 1)) (x (get-lexeme p-in) (get-lexeme p-in)))
        ((equal? x 'EOF) 'fini)
        (printf "~a : ~a~n" i x)))))

; (error "STOP AFTER LEXER")   ; <--- pour ne tester que le lexer !

(printf "Generating compiler...\n")

(define parse
  (parser
   [start go]
   [end EOF]
   
   [tokens value-tokens op-tokens]
   
   [error (lambda (a b c) (printf "ERROR with token : ~a\n" b))]
   
   [precs (nonassoc LOWER_THAN_ELSE)
          (nonassoc else)
          (nonassoc :=)
          ...]
   
   [grammar
    (go       [() 'fini]
              [(program) $1])
    (program  ...)
    ; j'ai environ 45 lignes dans ma grammaire...
    ...]))

(define (missp->scheme s)     ; string-missp -> pretty-string-scheme
  (let* ((p-in (open-input-string s))
         (scan (lambda () (get-lexeme p-in)))
         (string-port-out (open-output-string)))
    (pretty-print (parse scan) string-port-out)
    (get-output-string string-port-out)))

(printf "missp->scheme : (string-missp --> pretty-string-scheme) is available\n\n")
(tester "(missp->scheme \"foo {let x:=2 in 
                          {print(x-3); print(x)}}\")")

(define (test-file)
  (printf "TEST : compiling \"test.msa\" file...\n")
  (call-with-input-file "test.msa"                            ; compilation du fichier de test
    (lambda (p-in)
      (let ((scan (lambda () (get-lexeme p-in))))             ; un scanneur sur le port p-in
        (letrec ((one-bloc (lambda ()
                             (let ((prgm (parse scan)))       ; on compile une expression
                               (if (equal? prgm 'fini)
                                   'fini
                                   (begin (pretty-print prgm) (one-bloc)))))))
          (one-bloc))))))

(test-file)

(define (test-interactive)
  (printf "Test interactif ! Entrez un programme valide puis appuyez sur EOF\n")
  (let ((scan (lambda () (get-lexeme (current-input-port)))))
    (letrec ((one-bloc (lambda ()
                         (printf "? ")
                         (let ((prgm (parse scan)))
                           (if (equal? prgm 'fini)
                               'fini
                               (begin (pretty-print prgm) (one-bloc)))))))
      (one-bloc))))

(test-interactive)

